from logging import getLogger

from django.http import HttpResponseNotFound
from django.shortcuts import render

from .models import RedisServer
from .query import get_db_details
from .query import get_key_details

logger = getLogger(__name__)


def inspect_view(request, server: RedisServer):
    stats = server.stats
    conn = server.connection
    key_details = None
    database_details = None

    if stats['status'] == 'UP':
        if 'key' in request.GET:
            key = request.GET['key']
            db = request.GET.get('db', 0)
            page = request.GET.get('page', 1)
            key_details = get_key_details(conn, db, key, page)
        else:
            database_details = {name[2:]: data for name, data in conn.info().items() if name.startswith('db')}
            total_keys = 0
            for db, details in database_details.items():
                total_keys += details['keys']
            if total_keys < server.sampling_threshold:
                for db, details in database_details.items():
                    details.update(
                        get_db_details(server, db),
                        active=True,
                    )
            elif 'db' in request.GET:
                db = request.GET['db']
                if db in database_details:
                    database_details[db].update(
                        get_db_details(server, db),
                        active=True,
                    )
                else:
                    return HttpResponseNotFound("Unknown database.")

    return render(
        request,
        "redisboard/inspect.html",
        {
            'databases': database_details,
            'key_details': key_details,
            'original': server,
            'stats': stats,
            'app_label': 'redisboard',
        },
    )



# Generated by Django 2.2.28 on 2022-04-12 13:15

from django.db import migrations
from django.db import models

import redisboard.models


class Migration(migrations.Migration):

    dependencies = [
        ('redisboard', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='redisserver',
            name='url',
            field=models.CharField(
                help_text='<a href="https://www.iana.org/assignments/uri-schemes/prov/redis">IANA-compliant</a> URL. Examples: <pre>'
                'redis://[[username]:[password]]@localhost:6379/0\n'
                'rediss://[[username]:[password]]@localhost:6379/0\n'
                'unix://[[username]:[password]]@/path/to/socket.sock?db=0</pre>',
                max_length=250,
                null=True,
                validators=[redisboard.models.validate_url],
                verbose_name='URL',
            ),
        ),
    ]



from functools import partial
from logging import getLogger

from django.conf import settings
from django.core.paginator import Paginator
from redis.exceptions import ResponseError

from .utils import LazySlicingIterable
from .utils import maybe_text

logger = getLogger(__name__)

REDISBOARD_ITEMS_PER_PAGE = getattr(settings, 'REDISBOARD_ITEMS_PER_PAGE', 100)

LENGTH_GETTERS = {
    b'list': lambda conn, key: conn.llen(key),
    b'string': lambda conn, key: conn.strlen(key),
    b'set': lambda conn, key: conn.scard(key),
    b'zset': lambda conn, key: conn.zcount(key, '-inf', '+inf'),
    b'hash': lambda conn, key: conn.hlen(key),
}


def get_key_info(conn, key):
    try:
        obj_type = conn.type(key)
        length_getter = LENGTH_GETTERS.get(obj_type)
        if not length_getter:
            return {
                'type': 'none',
                'name': key,
                'length': "n/a",
                'error': "The key does not exist",
                'ttl': "n/a",
                'refcount': "n/a",
                'encoding': "n/a",
                'idletime': "n/a",
            }

        pipe = conn.pipeline()

        try:
            pipe.object('REFCOUNT', key)
            pipe.object('ENCODING', key)
            pipe.object('IDLETIME', key)
            length_getter(pipe, key)
            pipe.ttl(key)

            refcount, encoding, idletime, obj_length, obj_ttl = pipe.execute()
        except ResponseError as exc:
            logger.exception("Failed to get object info for key %r: %s", key, exc)
            return {
                'type': obj_type,
                'name': key,
                'length': "n/a",
                'error': str(exc),
                'ttl': "n/a",
                'refcount': "n/a",
                'encoding': "n/a",
                'idletime': "n/a",
            }
        return {
            'type': maybe_text(obj_type),
            'name': key,
            'length': obj_length,
            'ttl': obj_ttl,
            'refcount': refcount,
            'encoding': maybe_text(encoding),
            'idletime': idletime,
        }
    except ResponseError as exc:
        logger.exception("Failed to get details for key %r: %s", key, exc)
        return {
            'type': "n/a",
            'length': "n/a",
            'name': key,
            'error': str(exc),
            'ttl': "n/a",
            'refcount': "n/a",
            'encoding': "n/a",
            'idletime': "n/a",
        }


VALUE_GETTERS = {
    'list': lambda conn, key, start=0, end=-1: [(pos + start, val) for (pos, val) in enumerate(conn.lrange(key, start, end))],
    'string': lambda conn, key, *args: [('string', conn.get(key))],
    'set': lambda conn, key, *args: list(enumerate(conn.smembers(key))),
    'zset': lambda conn, key, start=0, end=-1: [(pos + start, val) for (pos, val) in enumerate(conn.zrange(key, start, end))],
    'hash': lambda conn, key, *args: conn.hgetall(key).items(),
    'n/a': lambda conn, key, *args: (),
    'none': lambda conn, key, *args: (),
}


def get_key_details(conn, db, key, page):
    conn.execute_command('SELECT', db)
    details = get_key_info(conn, key)
    details['db'] = db
    if details['type'] in ('list', 'zset'):
        details['data'] = Paginator(
            LazySlicingIterable(lambda: details['length'], partial(VALUE_GETTERS[details['type']], conn, key)), REDISBOARD_ITEMS_PER_PAGE
        ).page(page)
    else:
        details['data'] = VALUE_GETTERS[details['type']](conn, key)

    return details


def get_db_details(server, db):
    conn = server.connection
    conn.execute_command('SELECT', db)
    size = conn.dbsize()
    is_sampling = False

    if size > server.sampling_threshold:
        is_sampling = True
        pipe = conn.pipeline()
        for _ in range(server.sampling_size):  # noqa
            pipe.randomkey()
        keys = set(filter(None, pipe.execute()))
    else:
        keys = conn.keys()

    data = {key: get_key_info(conn, key) for key in keys}
    return {
        'data': data,
        'is_sampling': is_sampling,
    }



import re
from datetime import datetime
from datetime import timedelta

import redis
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from .utils import cached_property

REDISBOARD_DETAIL_FILTERS = [
    re.compile(name)
    for name in getattr(
        settings,
        'REDISBOARD_DETAIL_FILTERS',
        (
            'aof_enabled',
            'bgrewriteaof_in_progress',
            'bgsave_in_progress',
            'changes_since_last_save',
            'db.*',
            'last_save_time',
            'multiplexing_api',
            'total_commands_processed',
            'total_connections_received',
            'uptime_in_days',
            'uptime_in_seconds',
            'vm_enabled',
            'redis_version',
        ),
    )
]
REDISBOARD_DETAIL_TIMESTAMP_KEYS = getattr(settings, 'REDISBOARD_DETAIL_TIMESTAMP_KEYS', ('last_save_time',))
REDISBOARD_DETAIL_SECONDS_KEYS = getattr(settings, 'REDISBOARD_DETAIL_SECONDS_KEYS', ('uptime_in_seconds',))

REDISBOARD_SLOWLOG_LEN = getattr(settings, 'REDISBOARD_SLOWLOG_LEN', 10)

REDISBOARD_SOCKET_TIMEOUT = getattr(settings, 'REDISBOARD_SOCKET_TIMEOUT', None)
REDISBOARD_SOCKET_CONNECT_TIMEOUT = getattr(settings, 'REDISBOARD_SOCKET_CONNECT_TIMEOUT', None)
REDISBOARD_SOCKET_KEEPALIVE = getattr(settings, 'REDISBOARD_SOCKET_KEEPALIVE', None)
REDISBOARD_SOCKET_KEEPALIVE_OPTIONS = getattr(settings, 'REDISBOARD_SOCKET_KEEPALIVE_OPTIONS', None)


def prettify(key, value):
    if key in REDISBOARD_DETAIL_SECONDS_KEYS:
        return key, timedelta(seconds=value)
    elif key in REDISBOARD_DETAIL_TIMESTAMP_KEYS:
        return key, datetime.fromtimestamp(value)
    else:
        return key, value


def validate_url(value):
    try:
        redis.connection.parse_url(value)
    except Exception as exc:
        raise ValidationError(str(exc))


class RedisServer(models.Model):
    class Meta:
        verbose_name = _("Redis Server")
        verbose_name_plural = _("Redis Servers")
        permissions = (("can_inspect", "Can inspect redis servers"),)

    label = models.CharField(
        _('Label'),
        max_length=50,
        blank=True,
        null=True,
    )
    url = models.CharField(
        _("URL"),
        max_length=250,
        unique=True,
        help_text=_(
            '<a href="https://www.iana.org/assignments/uri-schemes/prov/redis">IANA-compliant</a> URL. Examples: <pre>'
            'redis://[[username]:[password]]@localhost:6379/0\n'
            'rediss://[[username]:[password]]@localhost:6379/0\n'
            'unix://[[username]:[password]]@/path/to/socket.sock?db=0</pre>'
        ),
        validators=[validate_url],
    )
    password = models.CharField(
        _("Password"), max_length=250, null=True, blank=True, help_text=_('You can also specify the password here (the field is masked).')
    )

    sampling_threshold = models.IntegerField(
        _("Sampling threshold"),
        default=1000,
        help_text=_("Number of keys after which only a sample (of random keys) is shown on the inspect page."),
    )
    sampling_size = models.IntegerField(
        _("Sampling size"),
        default=200,
        help_text=_("Number of random keys shown when sampling is used. Note that each key translates to a RANDOMKEY call in redis."),
    )

    @cached_property
    def connection(self) -> redis.StrictRedis:
        return redis.StrictRedis(
            single_connection_client=True,
            connection_pool=redis.ConnectionPool.from_url(
                self.url,
                password=self.password,
                retry_on_timeout=True,
            ),
        )

    @connection.deleter
    def connection(self, value):
        value.connection_pool.disconnect()

    @cached_property
    def stats(self):
        try:
            conn = self.connection
            info = conn.info()
            slowlog = conn.slowlog_get()
            slowlog_len = conn.slowlog_len()
            return {
                'status': 'UP',
                'details': info,
                'memory': "%s (peak: %s)" % (info['used_memory_human'], info.get('used_memory_peak_human', 'n/a')),
                'clients': info['connected_clients'],
                'brief_details': dict(prettify(k, v) for name in REDISBOARD_DETAIL_FILTERS for k, v in info.items() if name.match(k)),
                'slowlog': slowlog,
                'slowlog_len': slowlog_len,
            }
        except (redis.exceptions.ConnectionError, redis.exceptions.TimeoutError):
            return {
                'status': 'DOWN',
                'clients': 'n/a',
                'memory': 'n/a',
                'details': {},
                'brief_details': {},
                'slowlog': [],
                'slowlog_len': 0,
            }
        except redis.exceptions.RedisError as exc:
            return {
                'status': 'ERROR: %s' % exc.args,
                'clients': 'n/a',
                'memory': 'n/a',
                'details': {},
                'brief_details': {},
                'slowlog': [],
                'slowlog_len': 0,
            }

    def slowlog_html(self):
        commands = []
        for log in self.stats['slowlog']:
            command = log['command']
            if isinstance(command, bytes):
                command = repr(command)[2:-1]

            if len(command) > 255:
                command = f'{command:252}...'

            commands.append((log['duration'], command))
        commands.sort(reverse=True)
        if commands:
            output = ''.join(f'<tr><th>{duration / 1000.0:.1f}ms</th><td>{command}</td></tr>' for duration, command in commands)
            return mark_safe(f'<table><tr><th colspan="2">Total: {self.stats["slowlog_len"]} items</th></tr>{output}</table>')

        else:
            return 'n/a'

    def details_html(self):
        output = []
        brief_details = self.stats['brief_details']
        for k, v in brief_details.items():
            k = k.replace('_', ' ')
            if isinstance(v, dict):
                keys = v.keys()
                output.append(f'<tr><td colspan="2"><table><tr><th>{k}</th>')
                output.append(''.join(f'<th>{k}</th>' for k in keys))
                output.append('</tr><tr><td></td>')
                output.append(''.join(f'<td>{v[k]}</td>' for k in keys))
                output.append('</tr></table></td></tr>')
            else:
                output.append(f'<tr><th>{k}</th><td>{v}</td></tr>')
        if output:
            return mark_safe(f'<table>{"".join(output)}</table>')
        return 'n/a'

    def cpu_utilization_html(self):
        stats = self.stats
        if stats['status'] != 'UP':
            return 'n/a'

        data = (
            'used_cpu_sys',
            'used_cpu_sys_children',
            'used_cpu_user',
            'used_cpu_user_children',
        )
        data = dict((k, stats['details'][k]) for k in data)
        total_cpu = sum(data.values())
        uptime = stats['details']['uptime_in_seconds']
        data['cpu_utilization'] = '%.3f%%' % (total_cpu / uptime if uptime else 0)

        data = sorted(data.items())

        output = []
        for k, v in data:
            k = k.replace('_', ' ')
            output.append(f'<tr><th>{k}</th><td>{v}</td></tr>')

        if output:
            return mark_safe(f'<table>{"".join(output)}</table>')
        return 'n/a'

    def __str__(self):
        if self.label:
            return f'{self.label} ({self.url})'
        else:
            return self.url



# Generated by Django 2.2.28 on 2022-04-12 13:59

from django.db import migrations
from django.db import models

import redisboard.models


class Migration(migrations.Migration):

    dependencies = [
        ('redisboard', '0003_fill_url'),
    ]

    operations = [
        migrations.AlterField(
            model_name='redisserver',
            name='password',
            field=models.CharField(
                blank=True,
                help_text='You can also specify the password here (the field is masked).',
                max_length=250,
                null=True,
                verbose_name='Password',
            ),
        ),
        migrations.AlterField(
            model_name='redisserver',
            name='url',
            field=models.CharField(
                default='missing',
                help_text='<a href="https://www.iana.org/assignments/uri-schemes/prov/redis">IANA-compliant</a> URL. Examples: <pre>'
                'redis://[[username]:[password]]@localhost:6379/0\n'
                'rediss://[[username]:[password]]@localhost:6379/0\n'
                'unix://[[username]:[password]]@/path/to/socket.sock?db=0</pre>',
                max_length=250,
                unique=True,
                validators=[redisboard.models.validate_url],
                verbose_name='URL',
            ),
            preserve_default=False,
        ),
        migrations.AlterUniqueTogether(
            name='redisserver',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='redisserver',
            name='hostname',
        ),
        migrations.RemoveField(
            model_name='redisserver',
            name='port',
        ),
    ]



from functools import update_wrapper

from django.contrib import admin
from django.http import HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from .models import RedisServer
from .views import inspect_view


class RedisServerAdmin(admin.ModelAdmin):
    class Media:
        css = {'all': ('redisboard/admin.css',)}

    list_display = (
        '__str__',
        'status',
        'memory',
        'clients',
        'details',
        'cpu_utilization',
        'slowlog',
        'tools',
    )

    list_filter = ('label',)
    ordering = ('label', 'url')

    def details(self, obj):
        return obj.details_html()

    details.short_description = _('Details')

    def cpu_utilization(self, obj):
        return obj.cpu_utilization_html()

    cpu_utilization.short_description = _('CPU Utilization')

    def slowlog(self, obj):
        return obj.slowlog_html()

    slowlog.short_description = _('Slowlog')

    def status(self, obj):
        return obj.stats['status']

    status.short_description = _('Status')

    def memory(self, obj):
        return obj.stats['memory']

    memory.short_description = _('Memory')

    def clients(self, obj):
        return obj.stats['clients']

    clients.short_description = _('Clients')

    def tools(self, obj):
        return format_html('<a href="{}">{}</a>', reverse("admin:redisboard_redisserver_inspect", args=(obj.id,)), _("Inspect"))

    tools.short_description = _('Tools')

    def get_urls(self):
        urlpatterns = super(RedisServerAdmin, self).get_urls()
        from django.urls import path

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            return update_wrapper(wrapper, view)

        return [path('<int:server_id>/inspect/', wrap(self.inspect_view), name='redisboard_redisserver_inspect')] + urlpatterns

    def inspect_view(self, request, server_id):
        server = get_object_or_404(RedisServer, id=server_id)
        if self.has_change_permission(request, server) and request.user.has_perm('redisboard.can_inspect'):
            return inspect_view(request, server)
        else:
            return HttpResponseForbidden("You can't inspect this server.")


admin.site.register(RedisServer, RedisServerAdmin)



import os
import re
import time

import psutil as psutil
import pytest
import requests
from process_tests import TestProcess
from process_tests import dump_on_error
from process_tests import wait_for_strings

from redisboard.models import RedisServer

TIMEOUT = int(os.getenv('TEST_TIMEOUT', 10))


@pytest.fixture
def redis_model(redis_server, db):
    server = RedisServer.objects.create(
        url=f'unix:///{redis_server}',
    )
    c = server.connection
    c.set('str', 'bar')
    c.sadd('set', 'bar')
    c.sadd('set', 'foo')
    c.hset('hash', 'key', 'val')
    c.hset('hash', 'foo', 'bar')
    c.lpush('list', 'foo', 'bar', 'foobar')
    c.zadd('sorted-set', {'foo': 1, 'b': 2})
    c.eval(
        """
        local c = 0
        while c < 1000000 do
            redis.call('get', 'str')
            c = c + 1
        end
    """,
        0,
    )
    yield server


@pytest.fixture
def empty_redis_model(redis_server, db):
    return RedisServer.objects.create(
        url=f'unix:///{redis_server}?db=1',
    )


@pytest.mark.django_db
def test_changelist(admin_client, redis_model):
    response = admin_client.get('/redisboard/redisserver/')
    print(response)
    content = response.content.decode('utf-8')
    assert '>UP</' in content
    assert '>redis version</' in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/">Inspect</a>'.format(redis_model.pk) in content


@pytest.mark.django_db
def test_inspect(admin_client, redis_model):
    response = admin_client.get('/redisboard/redisserver/{:d}/inspect/'.format(redis_model.pk))
    print(response)
    content = response.content.decode('utf-8')
    assert 'Database 0:' in content
    assert '5 keys' in content
    assert '>UP</' in content
    assert '>redis version</' in content

    assert '<a href="/redisboard/redisserver/{:d}/inspect/?key=set&db=0">set</a>'.format(redis_model.pk) in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/?key=str&db=0">str</a>'.format(redis_model.pk) in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/?key=hash&db=0">hash</a>'.format(redis_model.pk) in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/?key=list&db=0">list</a>'.format(redis_model.pk) in content
    assert '<a href="/redisboard/redisserver/{:d}/inspect/?key=sorted-set&db=0">sorted-set</a>'.format(redis_model.pk) in content


@pytest.mark.django_db
def test_inspect_empty(admin_client, empty_redis_model):
    response = admin_client.get('/redisboard/redisserver/{:d}/inspect/'.format(empty_redis_model.pk))
    print(response)
    content = response.content.decode('utf-8')
    assert 'No databases found' in content
    assert '>UP</' in content
    assert '>redis version</' in content


@pytest.mark.django_db
@pytest.mark.parametrize("key", ["set", "str", "hash", "list", "sorted-set"])
def test_key_details(admin_client, redis_model, key):
    response = admin_client.get('/redisboard/redisserver/{:d}/inspect/?key={}'.format(redis_model.pk, key))
    print(response)
    content = response.content.decode('utf-8')
    assert '<h2>Key data: {}</h2>'.format(key) in content


@pytest.mark.parametrize('entrypoint', ['redisboard', 'python -mredisboard'])
def test_cli(entrypoint, tmpdir):
    args = ['127.0.0.1:0', '--password', 'foobar', '--storage', str(tmpdir)]
    with TestProcess(*entrypoint.split() + args) as process:
        with dump_on_error(process.read):
            wait_for_strings(process.read, TIMEOUT, "server at http://")
            t = time.time()
            port = None
            while time.time() - t < TIMEOUT and port is None:
                for conn in psutil.Process(process.proc.pid).connections('all'):
                    print(conn)
                    if conn.status == psutil.CONN_LISTEN and conn.laddr[0] == '127.0.0.1':
                        port = conn.laddr[1]
                        break
            assert port
            with requests.Session() as session:
                resp = session.get('http://127.0.0.1:%s/' % port)
                (csrftoken,) = re.findall('name=[\'"]csrfmiddlewaretoken[\'"] value=[\'"](.*?)[\'"]', resp.text)
                resp = session.post(
                    'http://127.0.0.1:%s/login/?next=/redisboard/redisserver/' % port,
                    data={
                        'csrfmiddlewaretoken': csrftoken,
                        'username': 'redisboard',
                        'password': 'foobar',
                    },
                )
                assert '<a href="/redisboard/redisserver/1/inspect/"' in resp.text
